def mklbl(prefix, n):
    return ["%s%s" % (prefix, i) for i in range(n)]

miindex = pd.MultiIndex.from_product([mklbl('A', 4),
                                      mklbl('B', 2),
                                      mklbl('C', 4),
                                      mklbl('D', 2)])
index  =[' '.join(col).strip() for col in miindex.values]
micolumns = pd.MultiIndex.from_tuples([('a', 'foo'),
                                       ('a', 'bar'),
                                       ('b', 'foo'),
                                       ('b', 'bah')],
                                      names=['lvl0', 'lvl1'])
cols  =[' '.join(col).strip() for col in micolumns.values]

dfrc = pd.DataFrame(np.arange(len(miindex) * len(micolumns)).reshape((len(miindex),
                                                                    len(micolumns))),
                  index=miindex,
                  columns=micolumns).sort_index().sort_index(axis=1)
dfr = pd.DataFrame(np.arange(len(miindex) * len(micolumns)).reshape((len(miindex),
                                                                    len(micolumns))),
                  index=miindex,
                  columns=cols).sort_index().sort_index(axis=1)
dfc = pd.DataFrame(np.arange(len(miindex) * len(micolumns)).reshape((len(miindex),
                                                                    len(micolumns))),
                  index=index,
                  columns=micolumns).sort_index().sort_index(axis=1)






import random
import gzip
import base64

import simplejson as json
import pandas as pd

from copy import deepcopy as copy

from .util import Util


class Params:
    """
    """

    def __init__(self,
                 width=700,
                 height=300,
                 theme='ag-theme-fresh',
                 css_rules=None,
                 quick_filter=False,
                 export_csv=False,
                 export_excel=False,
                 implicit_col_defs=True,
                 index=True,
                 keep_multiindex=True,
                 grid_data=None,
                 grid_options=None,
                 grid_options_multi=None,
                 license=None,
                 hide_grid=False,
                 compress_data=False,
                 verbose=False,
                 **kwargs):
        """
        Set params incl. defaults
        """
        # attributes
        self.uuid = random.randint(0, int(1e9))
        self.width = width
        self.height = height + 20
        self.theme = theme
        self.css_rules = css_rules
        self.quick_filter = quick_filter
        self.export_csv = export_csv
        self.export_excel = export_excel
        self.implicit_col_defs = implicit_col_defs
        self.index = index
        self.keep_multiindex = keep_multiindex
        self.grid_data_in = copy(grid_data)
        self.grid_data = None
        self.grid_options = copy(grid_options)
        self.grid_options_multi = copy(grid_options_multi)
        self.license = license
        self.hide_grid = hide_grid
        self.compress_data = compress_data

        # if not self.widthIframe:
        #     self.widthIframe = self.width + 2 * self.borderPx
        # if not self.heightIframe:
        #     isTitle = self.titleText is not None
        #     self.heightIframe = self.height + 30 * isTitle + 50 + 2 * self.borderPx + 5

        self.valid = self.check(verbose=verbose)

        if self.css_rules is not None:
            self.css_rules = Util.build_css_rules(self.css_rules)

        if self.license is not None:
            self.license = Util.encode_b64(self.license)

        self.is_grid_options_multi = True if grid_options_multi is not None else False
        self.is_grid_options_multi = int(self.is_grid_options_multi)

        if self.is_grid_options_multi:
            grid_options_multi_2 = []
            for name, options in self.grid_options_multi:
                self.grid_data, options_2 = self.preprocess_input(
                    self.grid_data_in,
                    options,
                    index=self.index,
                    keep_multiindex=self.keep_multiindex,
                    verbose=verbose)
                grid_options_multi_2.append((name, options_2))
            self.grid_options_multi = grid_options_multi_2

        else:
            self.grid_data, self.grid_options = self.preprocess_input(
                self.grid_data_in,
                self.grid_options,
                index=self.index,
                keep_multiindex=self.keep_multiindex,
                verbose=verbose)

        self.grid_data_json = Util.build_data(self.grid_data)
        if self.compress_data:
            json_bytes = self.grid_data_json.encode('utf-8')
            compressed = gzip.compress(json_bytes,
                                       compresslevel=9)
            b64_bytes = base64.encodebytes(compressed)
            b64_str = b64_bytes.decode('utf-8')
            self.grid_data_json = b64_str

        if self.is_grid_options_multi:
            self.grid_options_multi_json = Util.build_options({'data': self.grid_options_multi})
            # self.grid_options_multi_json = Util.build_options(self.grid_options_multi)
            # print(self.grid_options_multi_json)
        else:
            self.grid_options_json = Util.build_options(self.grid_options)
            # print(self.grid_options_json)

    def preprocess_input(self,
                         grid_data,
                         grid_options,
                         index,
                         keep_multiindex,
                         verbose=False):
        """
        """
        if Util.is_multiindex_df(grid_data):
            grid_data_2, grid_options_2 = Util.prepare_multiindex_df(
                grid_data,
                grid_options,
                index=index,
                keep_multiindex=keep_multiindex,
                verbose=verbose)

        elif Util.is_df(grid_data):
            grid_data_2, grid_options_2 = Util.prepare_singleindex_df(
                grid_data,
                grid_options,
                index=index,
                verbose=verbose)

        else:
            grid_options_2 = grid_options

        grid_options_2 = Util.update_columnTypes(
            grid_options_2,
            verbose=verbose)

        return grid_data_2, grid_options_2

    def check(self, verbose=False):
        """
        """
        msg = 'width must be a number of pixels'
        if isinstance(self.width, str):
            msg = 'width must end with "px" or "%"'
            assert self.width.endswith('px') or self.width.endswith('%'), msg
        else:
            msg = 'width must be a string or an int'
            assert isinstance(self.width, int), msg

        msg = 'height must be a number of pixels'
        if isinstance(self.height, str):
            msg = 'height must end with "px" or "%"'
            assert self.height.endswith('px') or self.height.endswith('%'), msg
        else:
            msg = 'height must be a string or an int'
            assert isinstance(self.height, int), msg

        msg = 'quick_filter must be a boolean'
        assert isinstance(self.quick_filter, bool), msg

        msg = 'export_csv must be a boolean'
        assert isinstance(self.export_csv, bool), msg

        msg = 'export_excel must be a boolean'
        assert isinstance(self.export_excel, bool), msg

        msg = 'hide_grid must be a boolean'
        assert isinstance(self.hide_grid, bool), msg

        if self.css_rules is not None:
            msg = 'css_rules must be a string'
            assert isinstance(self.css_rules, str), msg

        li_theme = ['ag-theme-fresh',
                    'ag-theme-dark',
                    'ag-theme-blue',
                    'ag-theme-material',
                    'ag-theme-bootstrap']
        msg = 'theme must be one of {}'.format(li_theme)
        assert self.theme in li_theme, msg

        msg = 'grid_data must be a dict or a dataframe'
        assert isinstance(self.grid_data_in, (list, pd.core.frame.DataFrame)), msg
        if isinstance(self.grid_data_in, list):
            msg = 'each element of grid_data must be a dict'
            for e in self.grid_data_in:
                assert isinstance(e, dict), msg

        msg = 'both grid_options and grid_options_multi cannot be set'
        assert (self.grid_options is None) or (self.grid_options_multi is None), msg

        msg = 'one exactly of grid_options or grid_options_multi mut be set'
        assert not((self.grid_options is None) and (self.grid_options_multi is None)), msg

        if self.grid_options is not None:
            msg = 'grid_options must be a dict'
            assert isinstance(self.grid_options, dict), msg

        if self.grid_options_multi is not None:
            msg = 'grid_options_multi must be a list or tuple'
            assert isinstance(self.grid_options_multi, (list, tuple)), msg
            msg1 = 'each element of grid_options_multi must be a list or tuple of length 2'
            msg2 = 'in each grid_options_multi element of length 2, the first one must be a string'
            msg3 = 'in each grid_options_multi element of length 3, the second one must be a dict'
            for e in self.grid_options_multi:
                assert isinstance(e, (list, tuple)) and len(e) == 2, msg1
                assert isinstance(e[0], str), msg2
                assert isinstance(e[1], dict), msg3

        if verbose:
            print('data is valid')

        return True

    def to_dict(self):
        """
        """
        d = copy(self.__dict__)
        d = {k: v for k, v in d.items() if v is not None}
        return d

    def pprint(self, indent=2):
        """
        """
        d = json.dumps(self.to_dict(),
                       sort_keys=True,
                       indent=indent)
        print(d)

    def __repr__(self):
        return str(self.to_dict())




import os
import re
import collections
import base64

import datetime as dt
import pandas as pd
import simplejson as json


class Util:

    @staticmethod
    def json_serial(obj):
        """
        """
        if isinstance(obj, (dt.datetime, dt.date, pd.Timestamp)):
            return obj.isoformat()

        return obj

    @staticmethod
    def strip_comments(code):
        lines = code.split('\n')
        lines = [e.strip() for e in lines]
        lines = [e for e in lines if not e.startswith('//')]
        code = '\n'.join(lines)
        return code

    @staticmethod
    def sanitize_str(string):
        string2 = Util.strip_comments(string)
        string2 = string2.replace('\n', '')
        string2 = string2.replace('\t', ' ')
        string2 = string2.replace('\"', '\'')
        return string2

    @staticmethod
    def sanitize_struct(e):
        # print('\nNEW')
        if isinstance(e, (list, tuple)):
            # print('list, tuple')
            return [Util.sanitize_struct(sub_e) for sub_e in e]
        elif isinstance(e, dict):
            # print('dict')
            return {k: Util.sanitize_struct(v) for k, v in e.items()}
        elif isinstance(e, str):
            # print('str')
            # print(e)
            return Util.sanitize_str(e)
        else:
            # print('native')
            # print(e)
            return e

    @staticmethod
    def is_df(data):
        if isinstance(data, pd.core.frame.DataFrame):
            return True
        return False

    @staticmethod
    def is_multiindex_row_df(df):
        if Util.is_df(df):
            if isinstance(df.index, pd.core.indexes.multi.MultiIndex):
                return True
        return False

    @staticmethod
    def is_multiindex_col_df(df):
        if isinstance(df, pd.core.frame.DataFrame):
            if isinstance(df.columns, pd.core.indexes.multi.MultiIndex):
                return True
        return False

    @staticmethod
    def is_multiindex_df(df):
        if Util.is_multiindex_row_df(df) or Util.is_multiindex_col_df(df):
            return True
        return False

    @staticmethod
    def build_colDefs_for_si_cols(df, verbose=False):
        colDefs = []
        for c in df.columns:
            dic = {}
            col = df[c]
            field = col.name
            header_name = field.title()
            if col.dtype.kind in 'O':
                # string
                dic['field'] = field
                dic['type'] = 'textColumn'
                dic['headerName'] = header_name
            if col.dtype.kind in 'ifc':
                # number
                dic['field'] = field
                dic['type'] = 'numberColumn'
                dic['headerName'] = header_name
            if col.dtype.kind in 'M':
                # date
                dic['field'] = field
                dic['type'] = 'dateColumn'
                dic['headerName'] = header_name
            colDefs.append(dic)
        return colDefs

    @staticmethod
    def build_colDefs_for_mi_cols(df):
        """
        create agGrid columnDefs dict for column grouping
        from multiindex dataframe columns
        """
        # utility
        def get_idx(s, x):
            li_headerName = [e['colName'] for e in s]
            if x not in li_headerName:
                return -1
            else:
                return li_headerName.index(x)

        mindexcol = df.columns
        li_idx_col = mindexcol.tolist()
        s = []
        for levels in li_idx_col:
            col = df.loc[:, levels]
            L = len(levels)
            s2 = s
            flat_field = None
            for k, e in enumerate(levels):
                if flat_field:
                    flat_field = flat_field + '-' + e
                else:
                    flat_field = e
                if k < L - 1:
                    i = get_idx(s2, e)
                    if i < 0:
                        new_e = {'colName': e,
                                 'headerName': e.title(),
                                 'children': []}
                        s2.append(new_e)
                        i = len(s2) - 1
                    s2 = s2[i]['children']
                else:
                    flat_field = flat_field.replace('.', '_')
                    new_e = {'field': flat_field,
                             'headerName': e.title()}
                    if col.dtype.kind in 'O':
                        # string
                        new_e['type'] = 'textColumn'
                    if col.dtype.kind in 'ifc':
                        # number
                        new_e['type'] = 'numberColumn'
                    if col.dtype.kind in 'M':
                        # date
                        new_e['type'] = 'dateColumn'
                    s2.append(new_e)
        return s

    @staticmethod
    def build_colDefs_for_mi_rows(df, keep_multiindex):
        """
        create agGrid columnDefs dict for column grouping
        from multiindex dataframe columns
        """
        mindexrow = df.index
        s = []
        for e in list(mindexrow.names):
            new_e = {'field': e,
                     'headerName': e.title(),
                     'rowGroup': True}
            if not keep_multiindex:
                new_e['hide'] = True
            s.append(new_e)
        return s

    @staticmethod
    def flatten_mi_col_df(dfmi):
        """
        create flattend dataframe
        multi index col ('a', 'b', 'c') turned to 'a-b-c'
        """
        df = dfmi.copy()
        cols = ['-'.join(col).strip() for col in df.columns.values]
        df.columns = cols
        df.columns.name = 'UniqueCol'
        return df

    @staticmethod
    def flatten_mi_row_df(dfmi):
        """
        create flattend dataframe
        multi index row added as regular column
        """
        df = dfmi.reset_index()
        return df

    @staticmethod
    def prepare_multiindex_df(dfmi,
                              options,
                              index=False,
                              keep_multiindex=False,
                              verbose=False):
        """
        Prepare multiindex dataframe (data) and options
        to display it with corresponding row grouping and
        column grouping
        To do that the dataframe is modified
        + multi index columns are flattened
        + multi index rows are made regular columns
        + columnDef in options are replaced with valid config
          (existing columnDefs if any is replaced)
        """

        df_data = dfmi

        if index:
            df_data = Util.add_index_as_col(df_data,
                                            verbose=verbose)

        if Util.is_multiindex_col_df(df_data):
            columnDefs_col = Util.build_colDefs_for_mi_cols(df_data)
            df_data = Util.flatten_mi_col_df(df_data)
        else:
            columnDefs_col = Util.build_colDefs_for_si_cols(df_data)

        if Util.is_multiindex_row_df(df_data):
            columnDefs_row = Util.build_colDefs_for_mi_rows(df_data,
                                                            keep_multiindex)
            df_data = Util.flatten_mi_row_df(df_data)
        else:
            columnDefs_row = []
            if index:
                df_data = Util.add_index_as_col(df_data,
                                                verbose=verbose)

        new_columnDefs = columnDefs_row + columnDefs_col

        options['columnDefs'] = new_columnDefs
        options['enableRowGroup'] = True

        return df_data, options

    @staticmethod
    def prepare_singleindex_df(data,
                               options,
                               index=False,
                               verbose=False):
        """
        Prepare single index dataframe (data) and options
        To do that the dataframe is modified
        + dots are replaced by underscore in column names
        + index is added in columns if requested
        + types are inferred from column types
        """

        data = Util.correct_df_col_name(data,
                                        verbose=verbose)
        if index:
            data = Util.add_index_as_col(data,
                                         verbose=verbose)

        if 'columnDefs' in options:
            options = Util.update_columnDefs(data,
                                             options,
                                             verbose=verbose)
        else:
            options = Util.implicit_columnDefs(data,
                                               options,
                                               verbose=verbose)

        return data, options

    @staticmethod
    def correct_df_col_name(data, verbose=False):
        new_col = [e.replace('.', '_') for e in data.columns]
        new_col_diff = [data.columns[i] != new_col[i]
                        for i in range(len(data.columns))]
        if sum(new_col_diff) > 0:
            if verbose:
                print('In dataframe column names "." are replaced by "_".', end=' ')
                print('Make sure columDefs match.')
            data.columns = new_col

        return data

    @staticmethod
    def add_index_as_col(data, verbose=False):
        data = data.reset_index()
        return data

    @staticmethod
    def update_columnDefs(df, grid_options, verbose=False):
        colDefs = grid_options.get('columnDefs', [])
        for colDef in colDefs:
            field = colDef.get('field')
            if field:
                if field in df.columns:
                    dic = colDef
                    col = df[field]
                    if col.dtype.kind in 'O':
                        # string
                        dic['type'] = 'textColumn'
                    if col.dtype.kind in 'ifc':
                        # number
                        dic['type'] = 'numberColumn'
                    if col.dtype.kind in 'M':
                        # date
                        dic['type'] = 'dateColumn'
        grid_options['columnDefs'] = colDefs
        return grid_options

    @staticmethod
    def implicit_columnDefs(df, grid_options, verbose=False):
        colDefs = Util.build_colDefs_for_si_cols(df, verbose=verbose)
        grid_options['columnDefs'] = colDefs
        return grid_options

    @staticmethod
    def update_columnTypes(grid_options, verbose=False):
        columnTypes = grid_options.get('columnTypes', {})

        numberColumn = {
            'filter': 'agNumberColumnFilter'
        }
        if not 'numberColumn' in columnTypes:
            columnTypes['numberColumn'] = numberColumn

        textColumn = {
            'filter': 'agTextColumnFilter'
        }
        if not 'textColumn' in columnTypes:
            columnTypes['textColumn'] = textColumn

        dateColumn = {
            'valueFormatter': 'helpers.dateFormatter',
            'filter': 'agDateColumnFilter',
            'filterParams': {
                'comparator': 'helpers.compareDates'
            }
        }
        if not 'dateColumn' in columnTypes:
            columnTypes['dateColumn'] = dateColumn

        grid_options['columnTypes'] = columnTypes
        return grid_options

    @staticmethod
    def build_css_rules(css_rules):
        css_rules = re.findall(r'[^\{]+\{[^\}]*\}',
                               css_rules,
                               re.MULTILINE)
        css_rules = [Util.sanitize_str(e) for e in css_rules]
        return css_rules

    @staticmethod
    def get_license(filename='.ag_grid_license'):
        path = os.path.join(os.path.expanduser('~'), filename)
        with open(path, 'r') as f:
            license = f.read()
        return license

    @staticmethod
    def encode_b64(string):
        return base64.b64encode(bytes(string, 'utf-8')).decode('utf-8')

    @staticmethod
    def build_data(data):
        if data is None:
            return {}

        if isinstance(data, pd.core.frame.DataFrame):
            data = data.to_dict(orient='records')

        data_json = json.dumps(data,
                               default=Util.json_serial,
                               ignore_nan=True)

        return data_json

    @staticmethod
    def build_options(options):
        if options is None:
            return {}

        options = Util.sanitize_struct(options)

        options_json = json.dumps(options,
                                  default=Util.json_serial,
                                  ignore_nan=True)

        return options_json
